<?php

namespace Drupal\varnish_file_purge\Services;

use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\file\FileInterface;
use Drupal\varnish_file_purge\Form\VarnishFilePurgeConfiguration;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Pool;
use Psr\Log\LoggerInterface;

/**
 * Helper class which sends the actual purge requests to Varnish.
 */
class VarnishFilePurger {

  /**
   * The configuration of the varnish file purger module.
   *
   * @var \Drupal\Core\Config\Config|\Drupal\Core\Config\ImmutableConfig
   */
  protected Config|ImmutableConfig $config;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface|\Drupal\Core\Logger\LoggerChannelInterface
   */
  protected LoggerInterface|LoggerChannelInterface $logger;

  /**
   * Constructs a VarnishFilePurgerHelper object.
   *
   * @param \GuzzleHttp\ClientInterface $client
   *   The HTTP client.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $loggerChannelFactory
   *   The logger channel factory.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   */
  public function __construct(
    protected ClientInterface $client,
    protected ConfigFactoryInterface $configFactory,
    protected LoggerChannelFactoryInterface $loggerChannelFactory,
    protected EntityTypeManagerInterface $entityTypeManager,
  ) {
    $this->config = $this->configFactory->get(VarnishFilePurgeConfiguration::CONFIG_KEY);
    $this->logger = $this->loggerChannelFactory->get('varnish_file_purge');
  }

  /**
   * Check if the varnish host is configured.
   *
   * @return bool
   *   Whether the host has been configured or not.
   */
  public function isConfigured(): bool {
    if (empty($this->config->get('host'))) {
      $this->logger->warning("Unable to purge: No Varnish host configured.");

      return FALSE;
    }

    return TRUE;
  }

  /**
   * Send purge requests to Varnish to purge all absolute URLs of a file.
   *
   * @param \Drupal\file\FileInterface $file
   *   The file whose URLs need to be purged.
   */
  public function purge(FileInterface $file): void {
    // Skip files that are in the temporary file stream.
    if (str_starts_with($file->getFileUri(), 'temporary://')) {
      return;
    }

    // Retrieve configuration settings.
    global $base_url;
    $domains = $this->config->get('domains') ?? $base_url;
    $domain_list = explode(',', $domains);
    $varnish_base_uri = $this->getBaseUri();
    $request_method = $this->config->get('request_method');
    $debug_mode = $this->config->get('debug');
    // Generate the absolute file URL to purge.
    $file_url = $file->createFileUrl();

    $this->purgeUrl($file_url, $request_method, $domain_list, $varnish_base_uri, $debug_mode);

    if (!$this->config->get('purge_styles')) {
      return;
    }

    $mimeType = $file->getMimeType();

    if (in_array($mimeType, ['image/png', 'image/jpeg', 'image/jpg'])) {
      $image_style_urls = $this->getExistingImageDerivatives($file_url);

      foreach ($image_style_urls as $image_style_url) {
        $this->purgeUrl($image_style_url, $request_method, $domain_list, $varnish_base_uri, $debug_mode);
      }
    }
  }

  /**
   * Purges a URL.
   *
   * @param string $url
   *   The URL to be purged.
   * @param string $request_method
   *   The purge method, PURGE or BAN.
   * @param array $domain_list
   *   The list of domains to purge the URL on.
   * @param string $varnish_base_uri
   *   The Varnish base URI.
   * @param string $debug_mode
   *   Whether debug mode is enabled or not.
   */
  protected function purgeUrl(
    string $url,
    string $request_method,
    array $domain_list,
    string $varnish_base_uri,
    string $debug_mode,
  ): void {
    // Generate a generator function to yield asynchronous requests.
    $requests_generator = function () use ($url, $request_method, $domain_list, $varnish_base_uri, $debug_mode) {
      foreach ($domain_list as $domain) {
        // Options for the HTTP request, including the Host header.
        $options = [
          'base_uri' => $varnish_base_uri,
          'headers' => [
            'Host' => $domain,
          ],
        ];

        // Yield an asynchronous Guzzle request.
        yield function () use ($url, $request_method, $domain, $options, $debug_mode) {
          return $this->client->requestAsync($request_method, $url, $options)
            ->then(
            // Success callback: Log purging if debug mode is enabled.
              function () use ($url, $domain, $debug_mode) {
                if ($debug_mode) {
                  $this->logger->notice('URL purged: %url on domain: %domain', ['%url' => $url, '%domain' => $domain]);
                }
              },
              // Failure callback: Log the error.
              function ($reason) use ($url) {
                $message = $reason instanceof \Exception ? $reason->getMessage() : (string) $reason;
                $this->logger->error('URL not purged: %url. Error: %message', [
                  '%url' => $url,
                  '%message' => $message,
                ]);
              }
            );
        };
      }
    };

    // Use Guzzle's Pool to handle the asynchronous requests with concurrency.
    $pool = new Pool($this->client, $requests_generator(), [
      'concurrency' => 10,
    ]);

    // Wait for all requests to complete.
    $pool->promise()->wait();
  }

  /**
   * Returns all the URLs of the existing images styles of the file.
   *
   * @param string $file_uri
   *   The File URL.
   *
   * @return array
   *   A list of URLs.
   */
  protected function getExistingImageDerivatives(string $file_uri): array {
    // Initialize an array to store existing derivative URLs.
    $existingDerivatives = [];

    // Load all image styles.
    try {
      $imageStyles = $this->entityTypeManager->getStorage('image_style')->loadMultiple();

      /** @var \Drupal\image\ImageStyleInterface $style */
      foreach ($imageStyles as $style) {
        $derivativeUri = $style->buildUri($file_uri);
        if (file_exists($derivativeUri)) {
          // The derivative exists, so generate the URL.
          $existingDerivatives[] = $style->buildUrl($file_uri);
        }
      }
    }
    catch (InvalidPluginDefinitionException | PluginNotFoundException $exception) {
      $this->logger->error("File image derivatives not purged: $file_uri. Error: {$exception->getMessage()}");
    }

    return $existingDerivatives;
  }

  /**
   * Returns the base URI of the varnish connection.
   *
   * @return string
   *   The base URI of the varnish connection.
   */
  protected function getBaseUri(): string {
    $base_uri = $this->config->get('scheme') ?? 'http://';
    $base_uri .= $this->config->get('host') ?? '';

    if (!empty($port = $this->config->get('port'))) {
      $base_uri .= ':' . $port;
    }

    return $base_uri;
  }

}
